# -*- coding: utf-8 -*-

# Copyright 2017 Juan José Martín Miralles
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python modules
import io
import subprocess
try:
	# Python 3
	from itertools import zip_longest
except ImportError:
	# Python 2
	from itertools import izip_longest as zip_longest

# Third party modules
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt  # Attention: include the .use('agg') before importing pyplot: DISPLAY issues
import mplleaflet
import numpy as np
from PIL import Image
from tqdm import tqdm
import smopy

# Own modules
from trackanimation.tracking import DFTrack
from trackanimation.utils import TrackException


class AnimationTrack:

	def __init__(self, df_points, dpi=100, bg_map=True, map_transparency=0.5):
		if not isinstance(df_points, list):
			df_points = [df_points]

		self.fig, self.axarr = plt.subplots(len(df_points), 1, facecolor='0.05', dpi=dpi)
		if not isinstance(self.axarr, np.ndarray):
			self.axarr = [self.axarr]

		self.map = []
		self.track_df = DFTrack()
		for i in range(len(df_points)):
			df = df_points[i].getTracks()	
			df.df['Axes'] = i
			self.track_df = self.track_df.concat(df)

			trk_bounds = df.getBounds()
			min_lat = trk_bounds.min_latitude
			max_lat = trk_bounds.max_latitude
			min_lng = trk_bounds.min_longitude
			max_lng = trk_bounds.max_longitude
			if bg_map:
				self.map.append(smopy.Map((min_lat, min_lng, max_lat, max_lng)))
				self.axarr[i].imshow(self.map[i].img, aspect='auto', alpha=map_transparency)
			else:
				self.axarr[i].set_ylim([min_lat, max_lat])
				self.axarr[i].set_xlim([min_lng, max_lng])

			self.axarr[i].set_facecolor('0.05')
			self.axarr[i].tick_params(color='0.05', labelcolor='0.05')
			for spine in self.axarr[i].spines.values():
				spine.set_edgecolor('white')

		self.fig.tight_layout()
		plt.subplots_adjust(wspace=0.1, hspace=0.1)

		if 'VideoFrame' in self.track_df.df:
			self.track_df = self.track_df.sort(['VideoFrame', 'Axes', 'CodeRoute'])
		else:
			self.track_df = self.track_df.sort(['Axes', 'Date'])

		self.track_df.df = self.track_df.df.reset_index(drop=True)


	def computePoints(self, track_df=None, linewidth=0.5):
		track_points = {}

		if track_df is None:
			points = self.track_df.toDict()
		else:
			points = track_df.toDict()

		for point, next_point in zip_longest(tqdm(points, desc='Computing points'), points[1:], fillvalue=None):
			track_code = str(point['CodeRoute']) + "_" + str(point['Axes'])

			# Check if the track is in the data structure
			if track_code in track_points:
				position = track_points[track_code]

				if len(position['lat']) > 1 and len(position['lng']) > 1:
					del position['lat'][0]
					del position['lng'][0]
					
				# Remove plotted line
				# for axarr in self.axarr:
				# 	for c in axarr.get_lines():
				# 		if c.get_gid() == track_code:
				# 			c.remove()
			else:
				position = {}
				position['lat'] = []
				position['lng'] = []

			lat = point['Latitude']
			lng = point['Longitude']
			if self.map:
				lng, lat = self.map[int(point['Axes'])].to_pixels(lat, lng)

			position['lat'].append(lat)
			position['lng'].append(lng)
			track_points[track_code] = position

			if 'Color' in point:
				self.axarr[int(point['Axes'])].plot(position['lng'], position['lat'], color=point['Color'], lw=linewidth, alpha=1)
			else:
				self.axarr[int(point['Axes'])].plot(position['lng'], position['lat'], color='deepskyblue', lw=linewidth, alpha=1)

			yield point, next_point


	def computeTracks(self, linewidth=0.5):
		df = self.track_df.getTracks().df

		df['track_code'] = df['CodeRoute'] + '_' + df['Axes'].map(str)
		grouped = df['track_code'].unique()

		for name in tqdm(grouped, desc='Groups'):
			df_slice = df[df['track_code'] == name]

			lat = df_slice['Latitude'].values
			lng = df_slice['Longitude'].values
			if self.map:
				lng, lat = self.map[int(df_slice['Axes'].unique())].to_pixels(lat, lng)

			self.axarr[int(df_slice['Axes'].unique())].plot(lng, lat, color='deepskyblue', lw=linewidth, alpha=1)


	def makeVideo(self, linewidth=0.5, output_file='video', framerate=5):
		cmdstring = ('ffmpeg',
			'-y',
			'-loglevel', 'quiet',
			'-framerate', str(framerate),
			'-f', 'image2pipe',
			'-i', 'pipe:',
			'-r', '25',
			'-s','1280x960',
			'-pix_fmt','yuv420p',
			output_file + '.mp4'
		)

		pipe = subprocess.Popen(cmdstring, stdin=subprocess.PIPE)

		for axarr in self.axarr:
			axarr.lines = []

		for point, next_point in self.computePoints(linewidth=linewidth):
			if self.isNewFrame(point, next_point):
				buffer = io.BytesIO()
				canvas = plt.get_current_fig_manager().canvas
				canvas.draw()
				pil_image = Image.frombytes('RGB', canvas.get_width_height(), canvas.tostring_rgb())
				pil_image.save(buffer, 'PNG')
				buffer.seek(0)
				pipe.stdin.write(buffer.read())

		pipe.stdin.close()


	def makeMap(self, linewidth=2.5, output_file='map'):
		for axarr in self.axarr:
			axarr.lines = []

		if self.map:
			raise TrackException('Map background found in the figure', 'Remove it to create an interactive HTML map.')

		if 'Color' in self.track_df.df:
			for point, next_point in self.computePoints(linewidth=linewidth):
				pass
		else:
			self.computeTracks(linewidth=linewidth)

		mplleaflet.save_html(fig=self.fig, tiles='esri_aerial',	 fileobj=output_file + '.html')#, close_mpl=False) # Creating html map


	def makeImage(self, linewidth=0.5, output_file='image', framerate=5, save_fig_at=None):
		for axarr in self.axarr:
			axarr.lines = []

		frame = 1
		if save_fig_at is not None or 'Color' in self.track_df.df:
			if not isinstance(save_fig_at, list):
				# If it is not a list, make a list of one element
				save_fig_at = [save_fig_at]

			for point, next_point in self.computePoints(linewidth=linewidth):
				if self.isNewFrame(point, next_point):
					second = frame / framerate
					if second in save_fig_at:
						plt.savefig(output_file + '_' + str(second) + '.png', facecolor=self.fig.get_facecolor())
					frame = frame + 1
		else:
			self.computeTracks(linewidth=linewidth)

		plt.savefig(output_file + '.png', facecolor=self.fig.get_facecolor())



	def isNewFrame(self, point, next_point):
		if next_point is not None:
			if 'VideoFrame' in point:
				new_frame = point['VideoFrame'] != next_point['VideoFrame']
			else:
				new_frame = point['Date'] != next_point['Date']		
		else:
			new_frame = False

		return new_frame


# -*- coding: utf-8 -*-

# Copyright 2017 Juan José Martín Miralles
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python modules
import os

# Own modules
from trackanimation.tracking import ReadTrack
from trackanimation.tracking import TrackException


def readTrack(directory_or_file, files_to_read=None):
	read_track = ReadTrack(directory_or_file)

	if directory_or_file.lower().endswith(('.csv')):
		return read_track.readCSV()
	elif directory_or_file.lower().endswith(('.gpx')) or os.path.isdir(directory_or_file):
		return read_track.readGPX(files_to_read)
	else:
		raise TrackException('Must specify a valid file name', directory_or_file)


# -*- coding: utf-8 -*-

# Copyright 2017 Juan José Martín Miralles
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python modules
import glob
import os

# Third party modules
import gpxpy
from gpxpy.gpx import GPXBounds
import pandas as pd
from pandas import DataFrame
from tqdm import tqdm
import geopy

# Own modules
from trackanimation import utils as trk_utils
from trackanimation.utils import TrackException


class DFTrack:

	def __init__(self, df_points=None, columns=None):
		if df_points is None:
			self.df = DataFrame()

		if isinstance(df_points, pd.DataFrame):
			self.df = df_points
		else:
			if columns is None:
				columns = ['CodeRoute', 'Latitude', 'Longitude', 'Altitude', 'Date', 
								'Speed', 'TimeDifference', 'Distance', 'FileName']
			self.df = DataFrame(df_points, columns=columns)


	def export(self, filename='exported_file', export_format='csv'):
		"""
		Export a data frame of DFTrack to JSON or CSV.

		Parameters
		----------
		export_format: string
			Format to export: JSON or CSV
		filename: string
			Name of the exported file
		"""
		if export_format.lower() == 'json':
			self.df.reset_index().to_json(orient='records', path_or_buf=filename+'.json')
		elif export_format.lower() == 'csv':
			self.df.to_csv(path_or_buf=filename+'.csv')
		else:
			raise TrackException('Must specify a valid format to export', "'%s'" % export_format)


	def getTracks(self):
		"""
		Makes a copy of the DFTrack.

		Explanation: 
			http://stackoverflow.com/questions/27673231/why-should-i-make-a-copy-of-a-data-frame-in-pandas

		Returns
		-------
		copy: DFTrack
			The copy of DFTrack.
		"""
		return self.__class__(self.df.copy(), list(self.df))


	def sort(self, column_name):
		"""
		Sorts the data frame by the specified column.

		Parameters
		----------
		column_name: string
			Column name to sort

		Returns
		-------
		sort: DFTrack
			DFTrack sorted
		"""
		if isinstance(column_name, list):
			for column in column_name:
				if column not in self.df:
					raise TrackException('Column name not found', "'%s'" % column)
		else:
			if column_name not in self.df:
				raise TrackException('Column name not found', "'%s'" % column_name)

		return self.__class__(self.df.sort_values(column_name), list(self.df))


	def getTracksByPlace(self, place, timeout=10, only_points=True):
		"""
		Gets the points of the specified place searching in Google's API
		and, if it does not get anything, it tries with OpenStreetMap's API.

		Parameters
		----------
		place: string
			Place to get the points
		timeout: int
			Time, in seconds, to wait for the geocoding service to respond 
			before returning a None value.
		only_points: boolean
			True to retrieve only the points that cross a place. False to
			retrive all the points of the tracks that cross a place.

		Returns
		-------
		place: DFTrack
			A DFTrack with the points of the specified place or 
			None if anything is found.
		"""
		track_place = self.getTracksByPlaceGoogle(place, timeout=timeout, only_points=only_points)
		if track_place is not None:
			return track_place

		track_place = self.getTracksByPlaceOSM(place, timeout=timeout, only_points=only_points)
		if track_place is not None:
			return track_place

		return None


	def getTracksByPlaceGoogle(self, place, timeout=10, only_points=True):
		"""
		Gets the points of the specified place searching in Google's API.

		Parameters
		----------
		place: string
			Place to get the points
		timeout: int
			Time, in seconds, to wait for the geocoding service to respond 
			before returning a None value.
		only_points: boolean
			True to retrieve only the points that cross a place. False to
			retrive all the points of the tracks that cross a place.

		Returns
		-------
		place: DFTrack
			A DFTrack with the points of the specified place or 
			None if anything is found.
		"""
		try:
			geolocator = geopy.GoogleV3()
			location = geolocator.geocode(place, timeout=timeout)
		except geopy.exc.GeopyError as geo_error:
			return None

		southwest_lat = float(location.raw['geometry']['bounds']['southwest']['lat'])
		northeast_lat = float(location.raw['geometry']['bounds']['northeast']['lat'])
		southwest_lng = float(location.raw['geometry']['bounds']['southwest']['lng'])
		northeast_lng = float(location.raw['geometry']['bounds']['northeast']['lng'])

		df_place = self.df[(self.df['Latitude'] < northeast_lat) & (self.df['Longitude'] < northeast_lng) & 
				(self.df['Latitude'] > southwest_lat) & (self.df['Longitude'] > southwest_lng)]

		if only_points:
			return self.__class__(df_place)

		track_list = df_place['CodeRoute'].unique().tolist()
		return self.__class__(self.df[self.df['CodeRoute'].isin(track_list)])


	def getTracksByPlaceOSM(self, place, timeout=10, only_points=True):
		"""
		Gets the points of the specified place searching in OpenStreetMap's API.

		Parameters
		----------
		place: string
			Place to get the points
		timeout: int
			Time, in seconds, to wait for the geocoding service to respond 
			before returning a None value.
		only_points: boolean
			True to retrieve only the points that cross a place. False to
			retrive all the points of the tracks that cross a place.

		Returns
		-------
		place: DFTrack
			A DFTrack with the points of the specified place or 
			None if anything is found.
		"""
		try:
			geolocator = geopy.Nominatim()
			location = geolocator.geocode(place, timeout=timeout)
		except geopy.exc.GeopyError as geo_error:
			return None

		southwest_lat = float(location.raw['boundingbox'][0])
		northeast_lat = float(location.raw['boundingbox'][1])
		southwest_lng = float(location.raw['boundingbox'][2])
		northeast_lng = float(location.raw['boundingbox'][3])

		df_place = self.df[(self.df['Latitude'] < northeast_lat) & (self.df['Longitude'] < northeast_lng) & 
				(self.df['Latitude'] > southwest_lat) & (self.df['Longitude'] > southwest_lng)]

		if only_points:
			return self.__class__(df_place)

		track_list = df_place['CodeRoute'].unique().tolist()
		return self.__class__(self.df[self.df['CodeRoute'].isin(track_list)])


	def getTracksByDate(self, start=None, end=None, periods=None, freq='D'):
		"""
		Gets the points of the specified date range 
		using various combinations of parameters.

		2 of 'start', 'end', or 'periods' must be specified.

		Date format recommended: 'yyyy-mm-dd'

		Parameters
		----------
		start: date
			Date start period
		end: date
			Date end period
		periods: int
			Number of periods. If None, must specify 'start' and 'end'
		freq: string
			Frequency of the date range

		Returns
		-------
		df_date: DFTrack
			A DFTrack with the points of the specified date range.
		"""
		if trk_utils.isTimeFormat(start) or trk_utils.isTimeFormat(end):
			raise TrackException('Must specify an appropiate date format', 'Time format found')

		rng = pd.date_range(start=start, end=end, periods=periods, freq=freq)
		
		df_date = self.df.copy()
		df_date['Date'] = pd.to_datetime(df_date['Date'])
		df_date['ShortDate'] = df_date['Date'].apply(lambda date: date.date().strftime('%Y-%m-%d'))
		df_date = df_date[df_date['ShortDate'].apply(lambda date: date in rng)]
		del df_date['ShortDate']

		df_date = df_date.reset_index(drop=True)

		return self.__class__(df_date, list(df_date))


	def getTracksByTime(self, start, end, include_start=True, include_end=True):
		"""
		Gets the points between the specified time range.

		Parameters
		----------
		start: datetime.time
			Time start period
		end: datetime.time
			Time end period
		include_start: boolean
		include_end: boolean
			
		Returns
		-------
		df_time: DFTrack
			A DFTrack with the points of the specified date and time periods.
		"""
		if not trk_utils.isTimeFormat(start) or not trk_utils.isTimeFormat(end):
			raise TrackException('Must specify an appropiate time format', trk_utils.TIME_FORMATS)

		df_time = self.df.copy()

		index = pd.DatetimeIndex(df_time['Date'])
		df_time = df_time.iloc[index.indexer_between_time(start_time=start, end_time=end, include_start=include_start, include_end=include_end)]

		df_time = df_time.reset_index(drop=True)

		return self.__class__(df_time, list(df_time))


	def pointVideoNormalize(self):
		df = self.df.copy()

		df_norm = pd.DataFrame()
		group_size = df.groupby('CodeRoute').size()
		max_value = group_size.max()
		name_max_value = group_size.idxmax()

		grouped = df['CodeRoute'].unique()

		for name in tqdm(grouped, desc='Groups'):
			df_slice = df[df['CodeRoute'] == name]
			df_slice = df_slice.reset_index(drop=True)
			div = int(max_value / len(df_slice)) + 1
			df_index = DataFrame(df_slice.index)
			df_slice['VideoFrame'] = df_index.apply(lambda x: x + 1 if name_max_value == name else x * div) 
			df_norm = pd.concat([df_norm, df_slice])

		df_norm = df_norm.reset_index(drop=True)

		return self.__class__(df_norm, list(df_norm))


	def timeVideoNormalize(self, time, framerate=5):
		df = self.df.copy()
		if time == 0:
			df['VideoFrame'] = 0
			df = df.reset_index(drop=True)
			return self.__class__(df, list(df))

		n_fps = time*framerate
		df = df.sort_values('Date')
		df_cum = trk_utils.calculateCumTimeDiff(df)
		grouped = df_cum['CodeRoute'].unique()

		df_norm = pd.DataFrame()
		point_idx = 1

		for name in tqdm(grouped, desc='Groups'):
			df_slice = df_cum[df_cum['CodeRoute'] == name]
			time_diff = float((df_slice[['TimeDifference']].sum() / time) / framerate) # Track duration divided by time and framerate

			df_range = df_slice[df_slice['CumTimeDiff'] == 0]
			df_range = df_range.reset_index(drop=True)
			df_range['VideoFrame'] = 0
			df_norm = pd.concat([df_norm, df_range])

			for i in tqdm(range(1, n_fps+1), desc='Num FPS', leave=False):
				x_start = time_diff*(i-1)
				x_end = time_diff*i

				df_range = df_slice[(df_slice['CumTimeDiff'] > x_start) & (df_slice['CumTimeDiff'] <= x_end)]
				df_range = df_range.reset_index(drop=True)

				if df_range.empty:
					df_start = df_slice[df_slice['CumTimeDiff'] <= x_start].tail(1)
					df_end = df_slice[df_slice['CumTimeDiff'] > x_end].head(1)

					if not df_start.empty and not df_end.empty:
						df_middlePoint = trk_utils.getPointInTheMiddle(df_start, df_end, time_diff, point_idx)
						df_range = DataFrame(df_middlePoint, columns=list(df_cum))

					point_idx = point_idx + 1
				else:
					point_idx = 1

				df_range['VideoFrame'] = i
				df_norm = pd.concat([df_norm, df_range])
		df_norm = df_norm.reset_index(drop=True)

		return self.__class__(df_norm, list(df_norm))


	def setColors(self, column_name, individual_tracks=True):
		if column_name not in self.df:
			raise TrackException('Column name not found', "'%s'" % column_name)

		df = self.df.copy()

		df_colors = pd.DataFrame()

		if individual_tracks:
			grouped = df['CodeRoute'].unique()

			for name in grouped:
				df_slice = df[df['CodeRoute'] == name]
				df_slice = df_slice.reset_index(drop=True)

				min = df_slice[column_name].min()
				max = df_slice[column_name].max()

				df_slice['Color'] = df_slice[column_name].apply(trk_utils.rgb, minimum=min, maximum=max)
				df_colors = pd.concat([df_colors, df_slice])

			df_colors = df_colors.reset_index(drop=True)
			return self.__class__(df_colors, list(df_colors))
		else:
			min = df[column_name].min()
			max = df[column_name].max()

			df['Color'] = df[column_name].apply(trk_utils.rgb, minimum=min, maximum=max) 
			df = df.reset_index(drop=True)

			return self.__class__(df, list(df))
			

	def dropDuplicates(self):
		"""
		Drop points of the same track with the same Latitude and Longitude.
		"""
		return self.__class__(self.df.drop_duplicates(['CodeRoute', 'Latitude', 'Longitude']))


	def toDict(self):
		"""
		Convert de data frame to a dictionary 
		like [{column -> value}, ... , {column -> value}]
		"""
		return self.df.to_dict('records')


	def getBounds(self):
		"""
		Get the bounds of the DFTrack
			
		Returns
		-------
		bounds: gpxpy.GPXBounds
		"""
		min_lat = self.df['Latitude'].min()
		max_lat = self.df['Latitude'].max()
		min_lng = self.df['Longitude'].min()
		max_lng = self.df['Longitude'].max()

		return GPXBounds(min_lat, max_lat, min_lng, max_lng)


	def concat(self, dfTrack):
		"""
		Concatenate DFTrack objects with 'self'

		Parameters
		----------
		dfTrack: DFTrack or list of DFTrack
			The ones that will be joined with 'self'
			
		Returns
		-------
		df_concat: DFTrack
			A DFTrack with the all the DFTrack concatenated
		"""
		if not isinstance(dfTrack, list):
			# If it is not a list of DFTrack, make a list of one element
			dfTrack = [dfTrack]

		df_concat = [self.df]  # First element is 'self'

		# From list of 'dfTrack', create a list of their dataframes
		for df in dfTrack:
			if not isinstance(df, DFTrack):
				raise TrackException("Parameter must be a 'DFTrack' object", '%s found' % type(df))

			df_concat.append(df.df) 

		return self.__class__(pd.concat(df_concat))


class ReadTrack:

	def __init__(self, directory_or_file):
		self.directory_or_file = directory_or_file
		self.points_list = []


	def readGPXFile(self, filename):
		try:
			with open(filename, "r") as f:
				prev_point = None
				head, tail = os.path.split(filename)
				code_route = tail.replace(".gpx", "")
				try:
					gpx = gpxpy.parse(f)
					for point in gpx.walk(only_points=True):
						speed = point.speed_between(prev_point)
						if speed is None:
							speed = 0

						timeDifference = point.time_difference(prev_point)
						if timeDifference is None:
							timeDifference = 0

						distance = point.distance_3d(prev_point)
						if not distance:
							distance = point.distance_2d(prev_point)
						if distance is None:
							distance = 0

						self.points_list.append([code_route, point.latitude, point.longitude, point.elevation, 
							point.time, speed, timeDifference, distance, gpx.name])

						prev_point = point
				except Exception as e:
					raise TrackException('GPX file "' + filename + '" malformed', e)
		except FileNotFoundError as e:
			raise TrackException('GPX file "' + filename + '" not found', e)


	def readGPX(self, files_to_read=None):
		if self.directory_or_file.lower().endswith(('.gpx')):
			self.readGPXFile(self.directory_or_file)
		else:
			n_file_read = 1
			for file in tqdm(glob.glob(self.directory_or_file + "*.gpx"), desc='Reading files'):
				try:
					self.readGPXFile(file)
				except TrackException as e:
					pass

				if files_to_read == n_file_read:
					break
				n_file_read += 1

		return DFTrack(self.points_list)


	def readCSV(self):
		try:
			return DFTrack(pd.read_csv(self.directory_or_file, sep=',', header=0, index_col=0))
		except FileNotFoundError as e:
			raise TrackException('CSV file not found', e)


# -*- coding: utf-8 -*-

# Copyright 2017 Juan José Martín Miralles
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python modules
import math
import datetime
import datetime as dt
from datetime import datetime

# Third party modules
import geopy
import geopy.distance as geo_dist
import pandas as pd


TIME_FORMATS = ['%H:%M', '%H%M', '%I:%M%p', '%I%M%p', '%H:%M:%S', '%H%M%S', '%I:%M:%S%p', '%I%M%S%p']



class TrackException(Exception):
    """
    Generic exception for TrackAnimation
    """
    def __init__(self, msg, original_exception):
        super(TrackException, self).__init__(msg + (": %s" % original_exception))
        self.original_exception = original_exception




def getBearing(start_point, end_point):
	"""
	Calculates the bearing between two points.

	Parameters
	----------
	start_point: geopy.Point
	end_point: geopy.Point
		
	Returns
	-------
	point: int
		Bearing in degrees between the start and end points.
	"""
	start_lat = math.radians(start_point.latitude)
	start_lng = math.radians(start_point.longitude)
	end_lat = math.radians(end_point.latitude)
	end_lng = math.radians(end_point.longitude)

	d_lng = end_lng - start_lng
	if abs(d_lng) > math.pi:
		 if d_lng > 0.0:
			 d_lng = -(2.0 * math.pi - d_lng)
		 else:
			 d_lng = (2.0 * math.pi + d_lng)

	tan_start = math.tan(start_lat / 2.0 + math.pi / 4.0)
	tan_end = math.tan(end_lat / 2.0 + math.pi / 4.0)
	dPhi = math.log(tan_end / tan_start)
	bearing = (math.degrees(math.atan2(d_lng, dPhi)) + 360.0) % 360.0

	return bearing


def getCoordinates(start_point, end_point, distance_meters):
	"""
	Calculates the new coordinates between two points depending
	of the specified distance and the calculated bearing.

	Parameters
	----------
	start_point: geopy.Point
	end_point: geopy.Point
	distance_meters: float
		
	Returns
	-------
	point: geopy.Point
		A new point between the start and the end points.
	"""
	bearing = getBearing(start_point, end_point)

	distance_km = distance_meters / 1000
	d = geo_dist.VincentyDistance(kilometers=distance_km)
	destination = d.destination(point=start_point, bearing=bearing)

	return geopy.Point(destination.latitude, destination.longitude)


def getPointInTheMiddle(start_point, end_point, time_diff, point_idx):
	"""
	Calculates a new point between two points depending of the
	time difference between them and the point index.

	Parameters
	----------
	start_point: DataFrame
	end_point: DataFrame
	time_diff: float
	point_idx: int
		Point index between the start and the end points
		
	Returns
	-------
	point: list
		A new point between the start and the end points.
	"""
	time_proportion = (time_diff * point_idx) / end_point['TimeDifference'].item()

	distance_proportion = end_point['Distance'].item() * time_proportion
	time_diff_proportion = end_point['TimeDifference'].item() * time_proportion
	speed = distance_proportion / time_diff_proportion
	distance = time_diff * speed
	cum_time_diff = int(start_point['CumTimeDiff'].item() + time_diff_proportion)
	# date = datetime.strptime(start_point['Date'].item(), '%Y-%m-%d %H:%M:%S') + dt.timedelta(seconds=int(time_diff_proportion))
	date = pd.to_datetime(start_point['Date'].astype(str), format='%Y-%m-%d %H:%M:%S') + dt.timedelta(seconds=int(time_diff_proportion))
	altitude = (end_point['Altitude'].item() + start_point['Altitude'].item()) / 2
	name = start_point['CodeRoute'].item()

	geo_start = geopy.Point(start_point['Latitude'].item(), start_point['Longitude'].item())
	geo_end = geopy.Point(end_point['Latitude'].item(), end_point['Longitude'].item())
	middlePoint = getCoordinates(geo_start, geo_end, distance_proportion)

	df_middlePoint = ([[name, middlePoint.latitude, middlePoint.longitude, altitude, 
			date, speed, int(time_diff), distance, None, cum_time_diff]])

	return df_middlePoint


def rgb(value, minimum, maximum):
	"""
	Calculates an rgb color of a value depending of
	the minimum and maximum values.

	Parameters
	----------
	value: float or int
	minimum: float or int
	maximum: float or int
		
	Returns
	-------
	rgb: tuple
	"""
	value = float(value)
	minimum = float(minimum)
	maximum = float(maximum)

	if minimum == maximum:
		ratio = 0
	else:
		ratio = 2 * (value - minimum) / (maximum - minimum)

	b = int(max(0, 255 * (1 - ratio)))
	r = int(max(0, 255 * (ratio - 1)))
	g = 255 - b - r
	
	return r/255.0, g/255.0, b/255.0


def calculateCumTimeDiff(df):
	"""
	Calculates the cumulative of the time difference
	between points for each track of 'dfTrack'

	Parameters
	----------
	df: DataFrame
		
	Returns
	-------
	df_cum: DataFrame
	"""
	df = df.copy()

	df_cum = pd.DataFrame()
	grouped = df['CodeRoute'].unique()

	for name in grouped:
		df_slice = df[df['CodeRoute'] == name]
		df_slice = df_slice.reset_index(drop=True)
		df_slice['CumTimeDiff'] = df_slice['TimeDifference'].cumsum()
		df_cum = pd.concat([df_cum, df_slice])

	df_cum = df_cum.reset_index(drop=True)

	return df_cum


def isTimeFormat(time):
	"""
	Check if 'time' variable has the format of one
	of the 'time_formats'
	"""
	if time is None:
		return False

	for time_format in TIME_FORMATS:
		try:
			datetime.strptime(time, time_format)
			return True
		except ValueError:
			pass

	return False


# -*- coding: utf-8 -*-

# Copyright 2017 Juan José Martín Miralles
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from os import path
from setuptools import setup, find_packages


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='trackanimation',

    version='1.0.0',

    description='GPS Track Animation Library',
    long_description=long_description,

    url='https://github.com/JoanMartin/trackanimation',

    author='Juan José Martín',
    author_email='',

    license='Apache License, Version 2.0',


    classifiers=[
        "Development Status :: 5 - Production/Stable",

        "License :: OSI Approved :: Apache Software License",

        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Multimedia :: Video",

        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],

    keywords='GPS GPX animation map geospatial geopositioning',

    packages=['trackanimation', ],
)